local LIB, REVISION = "LibDisplayAssist-1.4", 1
if not LibStub then
	error(LIB .. " requires LibStub", 0)
end

local lib, oldRevision = LibStub:NewLibrary(LIB, REVISION)
if not lib then return end

local ERROR_EVENT = "bad argument #%d to %q (unsupported event: %s)"
local ERROR_RATIO = "bad argument #%d to %q ([0, 1] expected, got %d)"
local ERROR_TYPE = "bad argument #%d to %q (%s expected, got %s)"

local pcall, tonumber, type = pcall, tonumber, type

local DISPLAY_WIDTH, DISPLAY_HEIGHT

--[[----------------------------------------------------------------------------
Version bridge
------------------------------------------------------------------------------]]
local callbacks, frame

if oldRevision then
	callbacks, frame = lib.__void()
else
	local function OnSizeChanged(...)
		frame.OnSizeChanged(...)
	end

	callbacks, frame = { }, CreateFrame('Frame')

	UIParent:HookScript('OnSizeChanged', OnSizeChanged)
	WorldFrame:HookScript('OnSizeChanged', OnSizeChanged)
end

frame:UnregisterAllEvents()
frame:Hide()

function lib.__void()
	wipe(lib)
	return callbacks, frame
end

--[[----------------------------------------------------------------------------
Anchors / GetOffset
------------------------------------------------------------------------------]]
local anchorPoint = setmetatable({
	BOTTOMLEFT = 'BOTTOMLEFT',
	LEFT = 'LEFT',
	TOPLEFT = 'TOPLEFT',
	TOP = 'TOP',
	TOPRIGHT = 'TOPRIGHT',
	RIGHT = 'RIGHT',
	BOTTOMRIGHT = 'BOTTOMRIGHT',
	BOTTOM = 'BOTTOM',
	CENTER = 'CENTER'
}, {
	__index = function(self, index)
		if type(index) == 'string' then
			local uCase = index:upper()
			if rawget(self, uCase) then
				self[index] = uCase
				return uCase
			end
		end
		return 'BOTTOMLEFT'
	end
})

local getOffset = setmetatable({
	BOTTOMLEFT	= function(x, y, w, h) return x, y end,
	LEFT		= function(x, y, w, h) return x, y + h * 0.5 end,
	TOPLEFT		= function(x, y, w, h) return x, y + h end,
	TOP			= function(x, y, w, h) return x + w * 0.5, y + h end,
	TOPRIGHT	= function(x, y, w, h) return x + w, y + h end,
	RIGHT		= function(x, y, w, h) return x + w, y + h * 0.5 end,
	BOTTOMRIGHT	= function(x, y, w, h) return x + w, y end,
	BOTTOM		= function(x, y, w, h) return x + w * 0.5, y end,
	CENTER		= function(x, y, w, h) return x + w * 0.5, y + h * 0.5 end
}, {
	__index = function(self, index)
		return self[anchorPoint[index]]
	end
})

--[[----------------------------------------------------------------------------
Helpers
------------------------------------------------------------------------------]]
local function CalculateX(object, x)
	return x * DISPLAY_WIDTH / object:GetEffectiveScale()
end

local function CalculateXY(object, x, y)
	local scale = object:GetEffectiveScale()
	return x * DISPLAY_WIDTH / scale, y * DISPLAY_HEIGHT / scale
end

local function CalculateY(object, y)
	return y * DISPLAY_HEIGHT / object:GetEffectiveScale()
end

local function CalculateScaledX(object, x)
	return x * object:GetEffectiveScale() / DISPLAY_WIDTH
end

local function CalculateScaledXY(object, x, y)
	local scale = object:GetEffectiveScale()
	return x * scale / DISPLAY_WIDTH, y * scale / DISPLAY_HEIGHT
end

local function CalculateScaledY(object, y)
	return y * object:GetEffectiveScale() / DISPLAY_HEIGHT
end

local function GetScaledRect(object)
	local left, bottom, width, height = object:GetRect()
	local scale = object:GetEffectiveScale()
	return left * scale / DISPLAY_WIDTH, bottom * scale / DISPLAY_HEIGHT, width * scale / DISPLAY_WIDTH, height * scale / DISPLAY_HEIGHT
end

local function GetOffset(object, point, asPercent)
	local x, y = getOffset[point](object:GetRect())
	if asPercent then
		local scale = object:GetEffectiveScale()
		x, y = x * scale / DISPLAY_WIDTH, y * scale / DISPLAY_HEIGHT
	end
	return x, y
end

local function GetRelOffset(object, point, relObject, relPoint, asPercent)
	local x2, y2, x1, y1 = getOffset[point](GetScaledRect(object))
	if type(relObject) == 'string' then
		relObject = _G[relObject]
	end
	if relObject then
		x1, y1 = getOffset[relPoint or point](GetScaledRect(relObject))
	else
		x1, y1 = getOffset[relPoint or point](0, 0, 1, 1)
	end
	x2, y2 = x2 - x1, y2 - y1
	if not asPercent then
		x2, y2 = CalculateXY(object, x2, y2)
	end
	return x2, y2
end

local function GetScaledHeight(object)
	return object:GetHeight() * object:GetEffectiveScale() / DISPLAY_HEIGHT
end

local function GetScaledSize(object)
	local width, height = object:GetSize()
	local scale = object:GetEffectiveScale()
	return width * scale / DISPLAY_WIDTH, height * scale / DISPLAY_HEIGHT
end

local function GetScaledWidth(object)
	return object:GetWidth() * object:GetEffectiveScale() / DISPLAY_WIDTH
end

local function SetOffset(object, point, offsetX, offsetY, asPercent)
	if asPercent then
		local scale = object:GetEffectiveScale()
		offsetX, offsetY = offsetX * DISPLAY_WIDTH / scale, offsetY * DISPLAY_HEIGHT / scale
	end
	object:ClearAllPoints()
	object:SetPoint(anchorPoint[point], nil, 'BOTTOMLEFT', offsetX, offsetY)
end

local function SetRelOffset(object, point, relObject, relPoint, offsetX, offsetY, asPercent)
	if asPercent then
		local scale = object:GetEffectiveScale()
		offsetX, offsetY = offsetX * DISPLAY_WIDTH / scale, offsetY * DISPLAY_HEIGHT / scale
	end
	object:ClearAllPoints()
	object:SetPoint(anchorPoint[point], relObject, anchorPoint[relPoint or point], offsetX, offsetY)
end

local function SetScaledHeight(object, height)
	object:SetHeight(height * DISPLAY_HEIGHT / object:GetEffectiveScale())
end

local function SetScaledSize(object, width, height)
	local scale = object:GetEffectiveScale()
	object:SetSize(width * DISPLAY_WIDTH / scale, height * DISPLAY_HEIGHT / scale)
end

local function SetScaledWidth(object, width)
	object:SetWidth(width * DISPLAY_WIDTH / object:GetEffectiveScale())
end

--[[----------------------------------------------------------------------------
Process callbacks
------------------------------------------------------------------------------]]
local pseudoEvent = {
	UI_SIZE_CHANGED = UIParent,
	WORLD_SIZE_CHANGED = WorldFrame
}

callbacks['DISPLAY_SIZE_CHANGED'] = callbacks['DISPLAY_SIZE_CHANGED'] or { }
callbacks['PLAYER_LOGOUT'] = callbacks['PLAYER_LOGOUT'] or { }
callbacks[UIParent] = callbacks[UIParent] or { }
callbacks[WorldFrame] = callbacks[WorldFrame] or { }

local function UpdateDisplaySize()
	local width, height = GetPhysicalScreenSize()
	DISPLAY_WIDTH, DISPLAY_HEIGHT = width * 768 / height, 768
end

local function ProcessCallbacks(self, event, width, height)
	if self then
		UpdateDisplaySize()
		width, height = DISPLAY_WIDTH, DISPLAY_HEIGHT
	end
	for reference, callback in pairs(callbacks[event]) do
		local ok, err = pcall(callback, reference, event, width, height)
		if not ok and err then
			geterrorhandler()(err)
		end
	end
end

function frame.OnSizeChanged(self, width, height)
	ProcessCallbacks(nil, self, width, height)
end

frame:SetScript('OnEvent', ProcessCallbacks)
frame:RegisterEvent('DISPLAY_SIZE_CHANGED')
frame:RegisterEvent('PLAYER_LOGOUT')

UpdateDisplaySize()

--[[----------------------------------------------------------------------------

LIB.RegisterCallback(reference, event, callback)

	Register for a callback when an event occurs.

Input:

	reference	(table)		A unique identifier that is passed to the
				(string)	callback and also used to unregister.

	event		(string)	The name of the event that triggers the callback,
							which is also passed to the callback.

	callback	(function)	The function to be called when the event happens.

------------------------------------------------------------------------------]]
function lib.RegisterCallback(reference, event, callback)
	local refType = type(reference)
	if refType ~= 'string' and refType ~= 'table' then
		error(ERROR_TYPE:format(1, "RegisterCallback", "string or table", refType), 2)
	elseif type(event) ~= 'string' then
		error(ERROR_TYPE:format(2, "RegisterCallback", 'string', type(event)), 2)
	elseif not callbacks[pseudoEvent[event] or event] then
		error(ERROR_EVENT:format(2, "RegisterCallback", event), 2)
	elseif type(callback) ~= 'function' then
		error(ERROR_TYPE:format(3, "RegisterCallback", 'function', type(callback)), 2)
	end

	local callbacks = callbacks[pseudoEvent[event] or event]
	if not callbacks then
		error(("bad argument #%d to %q (invalid event)"):format(2, "RegisterCallback"), 2)
	end
	callbacks[reference] = callback
end

--[[----------------------------------------------------------------------------

LIB.UnregisterCallback(reference, event, callback)

	Unregister for a callback when an event occurs.

Input:

	reference	(table)		The unique identifier that was used to
				(string)	register the callback.

	event		(string)	The name of the event that triggers the callback.

------------------------------------------------------------------------------]]
function lib.UnregisterCallback(reference, event)
	local refType = type(reference)
	if refType ~= 'string' and refType ~= 'table' then
		error(ERROR_TYPE:format(1, "UnregisterCallback", "string or table", refType), 2)
	elseif type(event) ~= 'string' then
		error(ERROR_TYPE:format(2, "UnregisterCallback", 'string', type(event)), 2)
	elseif not callbacks[pseudoEvent[event] or event] then
		error(ERROR_EVENT:format(2, "UnregisterCallback", event), 2)
	end

	local callbacks = callbacks[pseudoEvent[event] or event]
	if not callbacks then
		error(("bad argument #%d to %q (invalid event)"):format(2, "UnregisterCallback"), 2)
	end
	callbacks[reference] = nil
end

--[[----------------------------------------------------------------------------

LIB.CalculateX(object, x)

	Given scaled component x, determine it's dimension relative to the width of
	the display.

Input:

	object	(table)		The object having it's diminsion determined.

	x		(number)	The horizontal component as a percentage of the display's
						width.

Output:

	x		(number)	The horizontal component.

------------------------------------------------------------------------------]]
function lib.CalculateX(object, x)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "CalculateX", 'table', type(object)), 2)
	elseif type(x) ~= 'number' then
		error(ERROR_TYPE:format(2, "CalculateX", 'number', type(x)), 2)
	elseif x < 0 or x > 1 then
		error(ERROR_RATIO:format(2, "CalculateX", x), 2)
	end
	return CalculateX(object, x)
end

lib._CalculateX = CalculateX

--[[----------------------------------------------------------------------------

LIB.CalculateXY(object, x, y)

	Given scaled components x and y, determine their dimensions relative to the
	size of the display.

Input:

	object	(table)		The object having it's diminsions determined.

	x		(number)	The horizontal component as a percentage of the
						display's width.

	y		(number)	The vertical component as a percentage of the display's
						height.

Output:

	x		(number)	The horizontal component.

	y		(number)	The vertical component.

------------------------------------------------------------------------------]]
function lib.CalculateXY(object, x, y)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "CalculateXY", 'table', type(object)), 2)
	elseif type(x) ~= 'number' then
		error(ERROR_TYPE:format(2, "CalculateXY", 'number', type(x)), 2)
	elseif x < 0 or x > 1 then
		error(ERROR_RATIO:format(2, "CalculateXY", x), 2)
	elseif type(y) ~= 'number' then
		error(ERROR_TYPE:format(3, "CalculateXY", 'number', type(y)), 2)
	elseif y < 0 or y > 1 then
		error(ERROR_RATIO:format(3, "CalculateXY", y), 2)
	end
	return CalculateXY(object, x, y)
end

lib._CalculateXY = CalculateXY

--[[----------------------------------------------------------------------------

LIB.CalculateY(object, y)

	Given scaled component y, determine it's dimension relative to the height of
	the display.

Input:

	object	(table)		The object having it's diminsion determined.

	y		(number)	The vertical component as a percentage of the display's
						height.

Output:

	y		(number)	The vertical component.

------------------------------------------------------------------------------]]
function lib.CalculateY(object, y)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "CalculateY", 'table', type(object)), 2)
	elseif type(y) ~= 'number' then
		error(ERROR_TYPE:format(2, "CalculateY", 'number', type(y)), 2)
	elseif y < 0 or y > 1 then
		error(ERROR_RATIO:format(2, "CalculateY", y), 2)
	end
	return CalculateY(object, y)
end

lib._CalculateY = CalculateY

--[[----------------------------------------------------------------------------

LIB.CalculateScaledX(object, x)

	Given component x, determine it's dimension as a percentage of the display's
	width.

Input:

	object	(table)		The object having it's diminsions determined.

	x		(number)	The horizontal component.

Output:

	x		(number)	The horizontal component as a percentage of the
						display's width.

------------------------------------------------------------------------------]]
function lib.CalculateScaledX(object, x)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "CalculateScaledX", 'table', type(object)), 2)
	elseif type(x) ~= 'number' then
		error(ERROR_TYPE:format(2, "CalculateScaledX", 'number', type(x)), 2)
	elseif x < 0 or x > 1 then
		error(ERROR_RATIO:format(2, "CalculateScaledX", x), 2)
	end
	return CalculateScaledX(object, x)
end

lib._CalculateScaledX = CalculateScaledX

--[[----------------------------------------------------------------------------

LIB.CalculateScaledXY(object, x, y)

	Given components x and y, determine their dimensions as percentages of the
	display's size.

Input:

	object	(table)		The object having it's diminsions determined.

	x		(number)	The horizontal component.

	y		(number)	The vertical component.

Output:

	x		(number)	The horizontal component as a percentage of the display's
						width.

	y		(number)	The vertical component as a percentage of the display's
						height.

------------------------------------------------------------------------------]]
function lib.CalculateScaledXY(object, x, y)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "CalculateScaledXY", 'table', type(object)), 2)
	elseif type(x) ~= 'number' then
		error(ERROR_TYPE:format(2, "CalculateScaledXY", 'number', type(x)), 2)
	elseif x < 0 or x > 1 then
		error(ERROR_RATIO:format(2, "CalculateScaledXY", x), 2)
	elseif type(y) ~= 'number' then
		error(ERROR_TYPE:format(3, "CalculateScaledXY", 'number', type(y)), 2)
	elseif y < 0 or y > 1 then
		error(ERROR_RATIO:format(3, "CalculateScaledXY", y), 2)
	end
	return CalculateScaledXY(object, x, y)
end

lib._CalculateScaledXY = CalculateScaledXY

--[[----------------------------------------------------------------------------

LIB.CalculateScaledY(object, y)

	Given component y, determine it's diminsion as a percentage of the display's
	height.

Input:

	object	(table)		The object having it's diminsion determined.

	y		(number)	The vertical component.

Output:

	y		(number)	The vertical component as a percentage of the display's
						height.

------------------------------------------------------------------------------]]
function lib.CalculateScaledY(object, y)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "CalculateScaledY", 'table', type(object)), 2)
	elseif type(y) ~= 'number' then
		error(ERROR_TYPE:format(2, "CalculateScaledY", 'number', type(y)), 2)
	elseif y < 0 or y > 1 then
		error(ERROR_RATIO:format(2, "CalculateScaledY", y), 2)
	end
	return CalculateScaledY(object, y)
end

lib._CalculateScaledY = CalculateScaledY

--[[----------------------------------------------------------------------------

LIB.GetDisplaySize()

	Get the display's current size.

Input:

	width	(number)	The horizontal size.

	height	(number)	The vertical size.

------------------------------------------------------------------------------]]
function lib.GetDisplaySize()
	return DISPLAY_WIDTH, DISPLAY_HEIGHT
end

--[[----------------------------------------------------------------------------

LIB.GetOffset(object, point, asPercent)

	Get an object's offset to one of it's anchor points from the display's
	bottom left corner.

Input:

	object		(table)		The object having it's offset determined.

	point		(string)	The anchor point that the offset will be calculated
							to, defaulting to BOTTOMLEFT.

	asPercent	(boolean)	Determines if the offset will instead be returned as
							a percentage of the display's size.

Output:

	offsetX		(number)	The horizontal offset.

	offsetY		(number)	The vertical offset.
	
------------------------------------------------------------------------------]]
function lib.GetOffset(object, point, asPercent)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "GetOffset", 'table', type(object)), 2)
	elseif point ~= nil and type(point) ~= 'string' then
		error(ERROR_TYPE:format(2, "GetOffset", 'string', type(point)), 2)
	elseif asPercent ~= nil and type(asPercent) ~= 'boolean' then
		error(ERROR_TYPE:format(3, "GetOffset", 'boolean', type(asPercent)), 2)
	end
	return GetOffset(object, point, asPercent)
end

lib._GetOffset = GetOffset

--[[----------------------------------------------------------------------------

LIB.GetRelOffset(object, point, relObject, relPoint, asPercent)

	Get an object's offset to one of it's anchor points from another object's
	anchor point.

Input:

	object		(table)		The object having it's offset determined.

	point		(string)	The anchor point that the offset will be calculated
							to, defaulting to BOTTOMLEFT.

	relObject	(string)	The object relative to which the original
				(table)		object is having it's offset determined,
							defaulting to the display.

	relPoint	(string)	The anchor point of relObject that the offset will be
							calculated to, defaulting to point.

	asPercent	(boolean)	Determines if the offset will instead be returned as
							a percentage of the display's size.

Output:

	offsetX		(number)	The horizontal offset.

	offsetY		(number)	The vertical offset.
	
------------------------------------------------------------------------------]]
function lib.GetRelOffset(object, point, relObject, relPoint, asPercent)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "GetRelOffset", 'table', type(object)), 2)
	elseif point ~= nil and type(point) ~= 'string' then
		error(ERROR_TYPE:format(2, "GetRelOffset", 'string', type(point)), 2)
	elseif relObject ~= nil and type(relObject) ~= 'table' then
		error(ERROR_TYPE:format(3, "GetRelOffset", 'table', type(relObject)), 2)
	elseif relPoint ~= nil and type(relPoint) ~= 'string' then
		error(ERROR_TYPE:format(4, "GetRelOffset", 'string', type(relPoint)), 2)
	elseif asPercent ~= nil and type(asPercent) ~= 'boolean' then
		error(ERROR_TYPE:format(5, "GetRelOffset", 'boolean', type(asPercent)), 2)
	end
	return GetRelOffset(object, point, relObject, relPoint, asPercent)
end

lib._GetRelOffset = GetRelOffset

--[[----------------------------------------------------------------------------

LIB.GetScaledHeight(object)

	Get an object's height relative to the size of the display.

Input:

	object		(table)		The object having it's size determined.

Output:

	height		(number)	The object's height as a percentage of the display's
							height.
	
------------------------------------------------------------------------------]]
function lib.GetScaledHeight(object)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "GetScaledHeight", 'table', type(object)), 2)
	end
	return GetScaledHeight(object)
end

lib._GetScaledHeight = GetScaledHeight

--[[----------------------------------------------------------------------------

LIB.GetScaledRect(object)

	Get an object's rectangle coordinates/diminsions relative to the size of the
	display.

Input:

	object		(table)		The object having it's rectangle coordinates/diminsions
							determined.

Output:

	left		(number)	The object's left coordinate as a percentage of the
							display's width.

	bottom		(number)	The object's bottom coordinate as a percentage of the
							display's height.

	width		(number)	The object's width as a percentage of the display's
							width.

	height		(number)	The object's height as a percentage of the display's
							height.
	
------------------------------------------------------------------------------]]
function lib.GetScaledRect(object)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "GetScaledRect", 'table', type(object)), 2)
	end
	return GetScaledRect(object)
end

lib._GetScaledRect = GetScaledRect

--[[----------------------------------------------------------------------------

LIB.GetScaledSize(object)

	Get an object's size relative to the size of the display.

Input:

	object		(table)		The object having it's size determined.

Output:

	width		(number)	The object's width as a percentage of the display's
							width.

	height		(number)	The object's height as a percentage of the display's
							height.
	
------------------------------------------------------------------------------]]
function lib.GetScaledSize(object)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "GetScaledSize", 'table', type(object)), 2)
	end
	return GetScaledSize(object)
end

lib._GetScaledSize = GetScaledSize

--[[----------------------------------------------------------------------------

LIB.GetScaledWidth(object)

	Get an object's width relative to the size of the display.

Input:

	object		(table)		The object having it's size determined.

Output:

	width		(number)	The object's width as a percentage of the display's
							width.

------------------------------------------------------------------------------]]
function lib.GetScaledWidth(object)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "GetScaledWidth", 'table', type(object)), 2)
	end
	return GetScaledWidth(object)
end

lib._GetScaledWidth = GetScaledWidth

--[[----------------------------------------------------------------------------

LIB.IsWideScreen()

	Determines if the current window size is wider than standard (4:3).

------------------------------------------------------------------------------]]
function lib.IsWideScreen()
	return DISPLAY_HEIGHT / DISPLAY_WIDTH < 0.75
end

--[[----------------------------------------------------------------------------

LIB.SetOffset(object, point, asPercent)

	Set an object's offset to an anchor point from the display's bottom left
	corner.

Input:

	object		(table)		The object having it's offset set.

	point		(string)	The anchor point that the offset will be set to,
							defaulting to BOTTOMLEFT.

	offsetX		(number)	The horizontal offset to be set.

	offsetY		(number)	The vertical offset to be set.

	asPercent	(boolean)	Determines if the offset is a percentage of the
							display's size.

------------------------------------------------------------------------------]]
function lib.SetOffset(object, point, offsetX, offsetY, asPercent)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "SetOffset", 'table', type(object)), 2)
	elseif point ~= nil and type(point) ~= 'string' then
		error(ERROR_TYPE:format(2, "SetOffset", 'string', type(point)), 2)
	elseif type(offsetX) ~= 'number' then
		error(ERROR_TYPE:format(3, "SetOffset", 'number', type(offsetX)), 2)
	elseif type(offsetY) ~= 'number' then
		error(ERROR_TYPE:format(4, "SetOffset", 'number', type(offsetY)), 2)
	elseif asPercent ~= nil and type(asPercent) ~= 'boolean' then
		error(ERROR_TYPE:format(5, "SetOffset", 'boolean', type(asPercent)), 2)
	end
	return SetOffset(object, point, offsetX, offsetY, asPercent)
end

lib._SetOffset = SetOffset

--[[----------------------------------------------------------------------------

LIB.SetRelOffset(object, point, asPercent)

	Set an object's offset to an anchor point from the display's bottom left
	corner.

Input:

	object		(table)		The object having it's offset set.

	point		(string)	The anchor point that the offset will be set to,
							defaulting to BOTTOMLEFT.

	relObject	(table)		The object relative to which the original
							object is having it's offset determined,
							defaulting to the display.

	relPoint	(string)	The anchor point of relObject that the offset will be
							calculated to, defaulting to point.

	offsetX		(number)	The horizontal offset to be set.

	offsetY		(number)	The vertical offset to be set.

	asPercent	(boolean)	Determines if the offset is a percentage of the
							display's size.

------------------------------------------------------------------------------]]
function lib.SetRelOffset(object, point, relObject, relPoint, offsetX, offsetY, asPercent)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "SetRelOffset", 'table', type(object)), 2)
	elseif point ~= nil and type(point) ~= 'string' then
		error(ERROR_TYPE:format(2, "SetRelOffset", 'string', type(point)), 2)
	elseif relObject ~= nil and type(relObject) ~= 'table' then
		error(ERROR_TYPE:format(3, "SetRelOffset", 'table', type(relObject)), 2)
	elseif relPoint ~= nil and type(relPoint) ~= 'string' then
		error(ERROR_TYPE:format(4, "SetRelOffset", 'string', type(relPoint)), 2)
	elseif type(offsetX) ~= 'number' then
		error(ERROR_TYPE:format(5, "SetRelOffset", 'number', type(offsetX)), 2)
	elseif type(offsetY) ~= 'number' then
		error(ERROR_TYPE:format(6, "SetRelOffset", 'number', type(offsetY)), 2)
	elseif asPercent ~= nil and type(asPercent) ~= 'boolean' then
		error(ERROR_TYPE:format(7, "SetRelOffset", 'boolean', type(asPercent)), 2)
	end
	return SetRelOffset(object, point, relObject, relPoint, offsetX, offsetY, asPercent)
end

lib._SetRelOffset = SetRelOffset

--[[----------------------------------------------------------------------------

LIB.SetScaledHeight(object, height)

	Set an object's height relative to the size of the display.

Input:

	object		(table)		The object having it's size determined.

	height		(number)	The height as a percentage of the display's height.

------------------------------------------------------------------------------]]
function lib.SetScaledHeight(object, height)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "SetScaledHeight", 'table', type(object)), 2)
	elseif type(height) ~= 'number' then
		error(ERROR_TYPE:format(2, "SetScaledHeight", 'number', type(height)), 2)
	elseif height < 0 or height > 1 then
		error(ERROR_RATIO:format(2, "SetScaledHeight", height), 2)
	end
	return SetScaledHeight(object, height)
end

lib._SetScaledHeight = SetScaledHeight

--[[----------------------------------------------------------------------------

LIB.SetScaledSize(object, width, height)

	Set an object's size relative to the size of the display.

Input:

	object		(table)		The object having it's size determined.

	width		(number)	The width as a percentage of the display's width.

	height		(number)	The height as a percentage of the display's height.

------------------------------------------------------------------------------]]
function lib.SetScaledSize(object, width, height)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "SetScaledSize", 'table', type(object)), 2)
	elseif type(width) ~= 'number' then
		error(ERROR_TYPE:format(2, "SetScaledSize", 'number', type(width)), 2)
	elseif width < 0 or width > 1 then
		error(ERROR_RATIO:format(2, "SetScaledSize", width), 2)
	elseif type(height) ~= 'number' then
		error(ERROR_TYPE:format(3, "SetScaledSize", 'number', type(height)), 2)
	elseif height < 0 or height > 1 then
		error(ERROR_RATIO:format(3, "SetScaledSize", height), 2)
	end
	return SetScaledSize(object, width, height)
end

lib._SetScaledSize = SetScaledSize

--[[----------------------------------------------------------------------------

LIB.SetScaledWidth(object, width)

	Set an object's width relative to the size of the display.

Input:

	object		(table)		The object having it's size determined.

	width		(number)	The width as a percentage of the display's width.

------------------------------------------------------------------------------]]
function lib.SetScaledWidth(object, width)
	if type(object) ~= 'table' then
		error(ERROR_TYPE:format(1, "SetScaledWidth", 'table', type(object)), 2)
	elseif type(width) ~= 'number' then
		error(ERROR_TYPE:format(2, "SetScaledWidth", 'number', type(width)), 2)
	elseif width < 0 or width > 1 then
		error(ERROR_RATIO:format(2, "SetScaledWidth", width), 2)
	end
	return SetScaledWidth(object, width)
end

lib._SetScaledWidth = SetScaledWidth

--[[----------------------------------------------------------------------------
LIB.AnchorPoints

The anchor points used for frames, as a dictionary.
------------------------------------------------------------------------------]]
if type(lib.AnchorPoints) ~= 'table' then
	lib.AnchorPoints = {
		BOTTOM = 'BOTTOM', BOTTOMLEFT = 'BOTTOMLEFT', BOTTOMRIGHT = 'BOTTOMRIGHT',
		CENTER = 'CENTER', LEFT = 'LEFT', RIGHT = 'RIGHT',
		TOP = 'TOP', TOPLEFT = 'TOPLEFT', TOPRIGHT = 'TOPRIGHT'
	}
end

--[[----------------------------------------------------------------------------
LIB.StrataLayers

The non-tooltip strata layers generally used by addons, as an array, from lowest
to highest.
------------------------------------------------------------------------------]]
if type(lib.StrataLayers) ~= 'table' then
	lib.StrataLayers = {
		'BACKGROUND', 'LOW', 'MEDIUM', 'HIGH', 'DIALOG', 'FULLSCREEN', 'FULLSCREEN_DIALOG'
	}
end
