local addonName, addon = ...

local min, pairs_iter, tonumber, GetMouseFocus = min, pairs(addon), tonumber, GetMouseFocus

local mt = addon.GetMetatable('Frame')

local SetFrameLevel, SetFrameStrata = mt.__index.SetFrameLevel, mt.__index.SetFrameStrata

local L, LDA, LSM = addon.L, LibStub('LibDisplayAssist-1.4'), LibStub('LibSharedMedia-3.0')
local CalculateXY, GetRelOffset, SetRelOffset = LDA._CalculateXY, LDA._GetRelOffset, LDA._SetRelOffset

local panels, strataLayers = addon.panels, LDA.StrataLayers

--[[----------------------------------------------------------------------------
Sections - Fix for new Backdrop/NineSlice issues (uses 'Center' index of panel)

sections[sectionName][panel]
------------------------------------------------------------------------------]]
local sections = setmetatable({ }, {
	__index = function(self, index)
		local value = setmetatable({ }, addon.mt_subtables)
		self[index] = value
		return value
	end
})

--[[----------------------------------------------------------------------------
Helpers
------------------------------------------------------------------------------]]
local anchors, backdropInsets = { Center = 'LEFT', Left = 'LEFT', Right = 'RIGHT' }, { left = 0, right = 0, top = 0, bottom = 0 }
local backdrop = { bgFile = "", tile = false, tileSize = 0, edgeFile = "", edgeSize = 0, insets = backdropInsets }
local connections, directions = { LEFT = 'RIGHT', RIGHT = 'LEFT' }, { Center = 1, Left = 1, Right = -1 }
local recycled = { }

local auxFrame, auxString, centerFrame, overlayFrame, stateDriver = { }, { }, { }, { }, { }

local Jostle = addon.Jostle

local function OnFadeFinished(panel)
	if panel.fadeInfo.endAlpha == 0 then
		panel:Hide()
	end
end

local function ApplyFade(panel, alpha)
	local fadeInfo = panel.fadeInfo
	fadeInfo.startAlpha = panel:GetAlpha()
	if alpha > fadeInfo.startAlpha then
		fadeInfo.mode = "IN"
		fadeInfo.timeToFade = panel.settings.alphaFadeIn * (alpha - fadeInfo.startAlpha)
	elseif alpha < fadeInfo.startAlpha then
		fadeInfo.mode = "OUT"
		fadeInfo.timeToFade = panel.settings.alphaFadeOut * (fadeInfo.startAlpha - alpha)
	else
		return
	end
	fadeInfo.endAlpha = alpha
	fadeInfo.finishedFunc = OnFadeFinished
	fadeInfo.finishedArg1 = panel
	fadeInfo.fadeTimer = 0
	UIFrameFade(panel, fadeInfo)
end

local function UpdateAlpha(panel, alpha)
	alpha = addon.hideConditions[alpha] or tonumber(alpha) or panel.settings.alphaBase
	if alpha < 0 then
		alpha = 0
	elseif alpha > 1 then
		alpha = 1
	end
	ApplyFade(panel, alpha)
end

--[[----------------------------------------------------------------------------
Scripts
------------------------------------------------------------------------------]]
local function OnLeave(self, motion)
	self:HideTooltip(motion)
	if self.settings.alphaMouse and not self:HasMouseFocus() then
		self:Unreveal()
	end
end

local function OnMouseDown(self, button)
	if button == 'LeftButton' and self.unlocked then
		self.x, self.y = self:GetCenter()
		self:HideTooltip()
		self:SetFrameStrata('TOOLTIP')
		self:StartMoving()
		self.moving = true
	end
end

local function OnMouseUp(self, button)
	if button == 'LeftButton' then
		if self.moving then
			self.moving = nil
			self:StopMovingOrSizing()
			self:SetFrameStrata(strataLayers[self.settings.strata])
			self:SaveOffset()
			self:ShowTooltip()
			Jostle.Refresh()
		end
	elseif button == 'RightButton' and addon.config and self.settings.rightClickConfig and GetMouseFocus() == self then
		addon.config[1]()														-- Do it this way to force loading of config module
		addon.PanelList:Select(self.id)
		self:ShowTooltip()
	end
end

local function OnEnter_AuxFrame(self)
	self.parent:Reveal()
end

--[[----------------------------------------------------------------------------
Methods
------------------------------------------------------------------------------]]
local methods = {
	["AdjustForWorldMapFrame"] = function(self, state)
		self:SetFrameStrata(state and 'BACKGROUND' or strataLayers[self.settings.strata])
	end,

	["AllChildren"] = function(self, method, ...)
		for section in pairs_iter, addon.sectionTypes, nil do
			local children = sections[section][self]
			for index = #children, 1, -1 do
				children[index][method](children[index], ...)
			end
		end
	end,

	["AnchorChildren"] = function(self, section)
		local children = sections[section][self]
		if children[1] then
			local anchor, frame, offset = anchors[section]
			local anchorTo = connections[anchor]
			if section ~= "Center" then
				frame, offset = self, self.settings["spacing" .. section .. "Edge"] * directions[section] * self.size
			else
				frame, offset = centerFrame[self], 0
			end

			local plugin = children[1]
			plugin:ClearAllPoints()
			plugin:SetPoint(anchor, frame, anchor, offset, 0)
			frame, offset = plugin, self.settings["spacing" .. section] * directions[section] * self.size

			for index = 2, #children do
				plugin = children[index]
				plugin:ClearAllPoints()
				plugin:SetPoint(anchor, frame, anchorTo, offset, 0)
				frame = plugin
			end

			if section == "Center" then
				addon.QueueMethod(self, "UpdateCenterWidth")
			end
		end
	end,

	['GetName'] = function(self)
		if self.settings.allowGlobals then
			return addon.GetPanelName(self.id)
		end
	end,

	["GetSection"] = function(self, section)
		if addon.sectionTypes[section] then
			return sections[section][self]
		end
	end,

	["GetSections"] = function(self)
		return sections['Left'][self], sections['Center'][self], sections['Right'][self]
	end,

	["HasMouseFocus"] = function(self)
		return self[GetMouseFocus()]
	end,

	["HideTooltip"] = function(self, motion)
		if not (motion and addon.CONFIG_IS_OPEN and GetMouseFocus() == self) then
			GameTooltip:Hide()
			if self.tooltipScale then
				GameTooltip:SetScale(self.tooltipScale)
				self.tooltipScale = nil
			end
		end
	end,

	["Recycle"] = function(self)
		if self.jostle then
			Jostle.Unregister(self)
		end
		stateDriver[self]:Recycle()
		self:Hide()
		self:AllChildren("Recycle")
		addon.PurgeQueue(self)
		auxFrame[self]:EnableMouse(false)
		auxFrame[self]:Hide()
		addon.tremove_byVal(addon, self)
		_G[addon.GetPanelName(self.id)] = nil
		panels[self.id], stateDriver[self] = nil, nil
		self[0], self = self[0], wipe(self)
		recycled[#recycled + 1] = self
	end,

	["Refresh"] = function(self)
		local settings = self.settings

		local width, height = CalculateXY(self, settings.width, settings.height)
		self.anchor, self.size = settings.anchor, height

		local level, strata = settings.level, strataLayers[settings.strata]
		self:SetFrameStrata(strata)
		self:SetFrameLevel(level)

		self:UpdateLockState()
		local stateDriver = stateDriver[self]
		if stateDriver:IsEnabled() then
			auxFrame[self]:EnableMouse(not self:HasMouseFocus() and settings.alphaMouse)
		end
		stateDriver:SetParameters(settings.alphaParameters, true)

		local clamp = settings.screenClamp
		self:SetClampedToScreen(false)
		self:SetClampRectInsets(clamp, -clamp, -clamp, clamp)
		self:SetSize(width, height)
		self:SetClampedToScreen(true)
		self:SetOffset(settings.offsetX, settings.offsetY)

		local fontSize, string = height * 0.85, auxString[self]
		string:SetSize(0, 0)
		string:SetFont([[Fonts\FRIZQT__.TTF]], fontSize > 1 and fontSize or 1)
		string:SetSize(min(width, string:GetStringWidth()), string:GetStringHeight())
		self:AllChildren("Refresh")

		local bgInset = settings.bgInset
		backdrop.bgFile = LSM:Fetch('statusbar', settings.bgTexture)
		backdrop.edgeFile = LSM:Fetch('border', settings.borderTexture) or LSM:Fetch('border', "None")
		backdrop.edgeSize = settings.borderSize
		for inset in pairs_iter, backdropInsets, nil do
			backdropInsets[inset] = bgInset
		end
		self.backdropInfo = nil													-- Force update
		self:SetBackdrop(backdrop)
		self:SetBackdropColor(settings.bgColorR, settings.bgColorG, settings.bgColorB, settings.bgColorA)
		self:SetBackdropBorderColor(settings.borderColorR, settings.borderColorG, settings.borderColorB, settings.borderColorA)

		local overlay = overlayFrame[self]
		overlay:SetTexture(LSM:Fetch('statusbar_overlay', settings.overlayTexture))
		if settings.overlayFlip then
			if settings.overlayFlop then
				overlay:SetTexCoord(1, 0, 1, 0)
			else
				overlay:SetTexCoord(1, 0, 0, 1)
			end
		else
			if settings.overlayFlop then
				overlay:SetTexCoord(0, 1, 1, 0)
			else
				overlay:SetTexCoord(0, 1, 0, 1)
			end
		end
		overlay:SetVertexColor(settings.overlayColorR, settings.overlayColorG, settings.overlayColorB, settings.overlayColorA)

		for section in pairs_iter, addon.sectionTypes, nil do
			self:AnchorChildren(section)
		end
	end,

	["SaveOffset"] = function(self)
		local settings = self.settings
		settings.offsetX, settings.offsetY = GetRelOffset(self, self.anchor, not self.jostle and UIParent or nil, nil, true)
	end,

	['SetFrameLevel'] = function(self, value)
		SetFrameLevel(self, value)
		SetFrameLevel(auxFrame[self], value + 3)
	end,

	['SetFrameStrata'] = function(self, value)
		SetFrameStrata(self, value)
		SetFrameStrata(auxFrame[self], value)
	end,

	["SetOffset"] = function(self, x, y)
		if self.settings.shiftUI then
			SetRelOffset(self, self.anchor, nil, nil, x, y, true)
			if self.jostle then
				Jostle.Refresh()
			else
				self.jostle = true
				Jostle.Register(self)
			end
		else
			SetRelOffset(self, self.anchor, UIParent, nil, x, y, true)
			if self.jostle then
				self.jostle = nil
				Jostle.Unregister(self)
			end
		end
	end,

	["ShowTooltip"] = function(self, motion)
		if self.unlocked or addon.CONFIG_IS_OPEN and not self.tooltipScale then
			addon.ReleaseTooltip()
			local settings, tooltip = self.settings, GameTooltip
			tooltip:SetOwner(self, 'ANCHOR_NONE')
			if not self.tooltipScale then										-- In case of multiple shows before a hide
				self.tooltipScale = addon.ConnectTooltip(self, tooltip, 0.8)
			end
			tooltip:SetText(addonName .. " - " .. (settings.alias or L["Panel"]), 1, 1, 1)
			if self.unlocked then
				tooltip:AddLine(L["Left click and hold to move this panel."])
			end
			if settings.rightClickConfig and (addon.CONFIG_IS_OPEN ~= addon.config[1] or addon.PanelList.selection ~= self.id) then
				tooltip:AddLine(L["Right click to change this panel's settings."])
			end
			tooltip:Show()
		end
	end,

	["UpdateCenterWidth"] = function(self)
		local center = sections['Center'][self]
		local left, right = center[1]:GetLeft(), center[#center]:GetRight()
		if left and right then
			centerFrame[self]:SetWidth(right - left + 1)
		else
			addon.QueueMethod(self, "UpdateCenterWidth")
		end
	end,

	["UpdateOrder"] = function(self)
		local settings = self.settings
		local order = (#strataLayers - settings.strata) * 1000 + (127 - settings.level)
		if self.order ~= order then
			self.order = order
			addon.tremove_byVal(addon, self)
			local insertPoint
			for index = 1, #addon do
				if order <= addon[index].order then
					insertPoint = index
					break
				end
			end
			tinsert(addon, insertPoint or #addon + 1, self)
		end		
	end,

-- WIP

	["UpdateLockState"] = function(self)
		local settings = self.settings
		if settings.lockPanel then
			if self.unlocked ~= false then
				self.unlocked = false
				auxFrame[self]:SetAlpha(0)
				self:Unreveal()
			end
		elseif self.unlocked ~= true then
			self.unlocked = true
			auxFrame[self]:SetAlpha(1)
			self:Reveal()
		end
	end,

	["Reveal"] = function(self)
		if self.revealed == 0 then
			self.revealed = 1
			stateDriver[self]:SetEnabled(false)
			auxFrame[self]:EnableMouse(false)
			ApplyFade(self, 1)
		else
			self.revealed = self.revealed + 1
		end
	end,

	["Unreveal"] = function(self)
		local revealed = self.revealed
		if revealed == 1 and self.unlocked ~= true then
			self.revealed = 0
			stateDriver[self]:SetEnabled(true)
			auxFrame[self]:EnableMouse(self.settings.alphaMouse)
		elseif revealed > 1 then
			self.revealed = revealed - 1
		end
	end,

	['EnableMouse'] = mt.__index.EnableMouse,
	['Hide'] = mt.__index.Hide,
	['SetAlpha'] = mt.__index.SetAlpha,
	['Show'] = mt.__index.Show
}

mt = { __index = setmetatable(methods, mt), __metatable = addonName }

--[[----------------------------------------------------------------------------
Global to addon
------------------------------------------------------------------------------]]
function addon.GetPanelName(id)
	return ("%s%s[%s]"):format(addonName, L["Panel"], id)
end

function addon.CreatePanel(id)
	if not id then
		id = addon.GenerateUniqueKey()
	end
	local self = recycled[#recycled]
	if self then
		recycled[#recycled] = nil
		if BackdropTemplateMixin then
			Mixin(self, BackdropTemplateMixin)									-- Recycle removes
		end
		self:SetClampedToScreen(false)
		auxFrame[self]:Show()
		self:Show()
	else
		self = CreateFrame('Frame', nil, UIParent, BackdropTemplateMixin and 'BackdropTemplate')
		self:EnableMouse(true)
		self:SetMovable(true)
		self:SetScript('OnEnter', methods.ShowTooltip)
		self:SetScript('OnLeave', OnLeave)
		self:SetScript('OnMouseDown', OnMouseDown)
		self:SetScript('OnMouseUp', OnMouseUp)

		local frame = CreateFrame('Frame', nil, UIParent)
		frame:SetAllPoints(self)
		frame:SetScript('OnEnter', OnEnter_AuxFrame)

		local texture = frame:CreateTexture(nil, 'BACKGROUND')
		texture:SetAllPoints()
		texture:SetTexture([[Interface\BUTTONS\WHITE8X8]])
		texture:SetVertexColor(0, 0.05, 1, 0.38196601125011)

		local string = frame:CreateFontString(nil, 'ARTWORK')
		string:SetPoint('CENTER')
		string:SetFont([[Fonts\FRIZQT__.TTF]], 12)
		string:SetJustifyH('CENTER')
		string:SetJustifyV('MIDDLE')
		string:SetNonSpaceWrap(false)
		string:SetShadowColor(0, 0, 0, 0.5)
		string:SetShadowOffset(1, -1)
		string:SetText(L["Unlocked"])
		string:SetTextColor(0.85, 0.85, 0.85)

		local overlay = self:CreateTexture(nil, 'OVERLAY')
		overlay:SetAllPoints()

		local center = CreateFrame('Frame', nil, self)
		center:SetPoint('CENTER')
		center:SetSize(1, 1)

		auxFrame[self], auxString[self], centerFrame[self], frame.parent, overlayFrame[self] = frame, string, center, self, overlay
		setmetatable(self, mt)
	end
	panels[id], self.id, self.size = self, id, 1

	self:SetSize(1, 1)
	self:SetAlpha(1)

	stateDriver[self] = LibStub('LibStateDriver-1.3').New(nil, 'SetObject', self, 'SetCallback', UpdateAlpha)
	self.fadeInfo = {
		mode = "IN",
		startAlpha = 0,
		endAlpha = 1,
		timeToFade = 0,
		finishedFunc = false,
		finishedArg1 = self,
		fadeTimer = 0
	}

	local settings = addon.GetPanelSettings(id)
	self.anchor, self.settings = settings.anchor, settings
	self:SetOffset(settings.offsetX, settings.offsetY)

	self:UpdateOrder()
	self.revealed = settings.lockPanel and 1 or 0								-- Prep for first call to UpdateLockState
	self:Refresh()																-- Calls UpdateLockState
	_G[addon.GetPanelName(id)] = settings.allowGlobals and self or nil

	local dataObj, plugins, pluginSettings, CreatePlugin = addon.dataObj, addon.plugins, addon.settings.plugins, addon.CreatePlugin
	for section in pairs_iter, addon.sectionTypes, nil do
		local children = settings[section]
		for index = 1, #children do
			local name = children[index]
			if dataObj[name] and not plugins[name] and pluginSettings[name].enable then
				CreatePlugin(name, dataObj[name])
			end
		end
	end

	addon.PanelList:Add(id)
	return self
end
